const logger = require('@firstfleet/fflogger');
const { errorLevels, slackChannels } = require("./constants");
const { handleError, logAndNotify, PostToSlack } = require("./handlers");
const { expressLogErrors, expressHandleErrors } = require("./middleware");
const errorTypes = require("./errorTypes");

const appName = process.env.PAPERTRAIL_PROGRAM || "UNKNOWN";

/** 
 * How long to wait before gracefully shutting down the server 
 * when we get an `uncaughtException` error 
 * */
const WAITBEFORESHUTDOWN = 1000;

/** 
 * Try to listen for Papertrail logging error events and post to Slack 
 * */
if (logger?.PaperTrailErrorEvent) {
    logger.PaperTrailErrorEvent.on('error', (error) => {
        try {
            console.error(error);
            logAndNotify(
                appName,
                "Error Handler Papertrail Init",
                `Unable to log to Papertrail: ${error.message}`,
                error.stack,
                slackChannels.WARNING
            );
        } catch (error) {
            console.error(error);
        }
    });
}

/**
 * Global handler for all uncaught exceptions,
 * gracefully restart the server after logging to slack
 */
process.on('uncaughtException', (error) => {
    try {
        console.error(error);
        logAndNotify(
            appName,
            'GLOBAL - uncaughtException',
            error.message,
            error.stack,
            slackChannels.CRITICAL
        );

        /** 
        * If logging has not finished in 1 second, go ahead and 
        * gracefully shutdown the server
        * */
        setTimeout(() => {
            process.exit(1);
        }, WAITBEFORESHUTDOWN)

    } catch (error) {
        console.error(error);

        /** 
         * We don't need a timeout here, as console.error is synchronous
         * */
        process.exit(1);
    }
});

/**
 * Global handler for all uncaught promise rejections.
 * If an uncaught promise rejection is thrown this will handle the error
 * gracefully.
 */
process.on('unhandledRejection', (error) => {
    try {
        console.error(error);
        logAndNotify(
            appName,
            'GLOBAL - unhandledRejection',
            error.message,
            error.stack,
            slackChannels.CRITICAL
        );
    } catch (error) {
        console.error(error);
    }
});

module.exports = {
    slackChannels,
    handleError,
    errorLevels,
    logAndNotify,
    PostToSlack,
    expressLogErrors,
    expressHandleErrors,
    errorTypes
};
